# -*- coding: utf-8 -*-
# Copyright 2016 Open Permissions Platform Coalition
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations under
# the License.
import importlib
import logging
import os
import pkgutil
from collections import defaultdict
from functools import wraps
from uuid import uuid4

from tornado.gen import coroutine

from .views import resource_version


__all__ = ['migration', 'Migration', 'collect', 'run_migrations']

MIGRATION_TEMPLATE = """
from __future__ import unicode_literals
import perch

VERSION = '{version}'
PREVIOUS_VERSION = '{previous_version}'


@perch.migrate.migration({resource_path}, VERSION, PREVIOUS_VERSION)
def migrate_{resource_type}(instance):
    # do stuff here

    return instance
"""


def migration_registry():
    return defaultdict(lambda: {
        'migrations': [],
        'subresources': migration_registry()
    })

_migrations = defaultdict(migration_registry)


class Migration(object):
    def __init__(self, func, resource, version, previous_version=''):
        self.func = func
        self.resource = resource
        self.previous_version = previous_version
        self.version = version

    def register(self):
        migrations = self._get_migrations()
        migrations['migrations'].append(self)

    def _get_migrations(self):
        if not hasattr(self.resource, 'parent_resources'):
            return _migrations[self.resource][self.previous_version]

        parents = self.resource.parent_resources()
        m = _migrations[parents[0]][self.previous_version]
        for p in parents[1:] + [self.resource]:
            m = m['subresources'][p]

        return m

    def __call__(self, instance):
        """
        Apply a migration

        :param instance: a perch.Document or perch.SubResource instance
        """
        logging.info('Migrating {} {}{} to {}'.format(
            instance.resource_type,
            instance.id,
            ' from ' + self.previous_version if self.previous_version else '',
            self.version
        ))

        return self.func(instance)

    def __unicode__(self):
        return "<Migrate {} {} to {}>".format(
            self.resource.resource_type, self.previous_version, self.version)

    def __str__(self):
        return self.__unicode__()

    def __repr__(self):
        return unicode(self)


def migration(resource, version, previous_version=''):
    """Register a migration function"""
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            migrated = func(*args, **kwargs)

            return migrated

        m = Migration(wrapper, resource, version, previous_version)
        m.register()

        return m

    return decorator


def create(resource_path, previous_version=None, package='perch.migrations'):
    """Create a new migration"""
    pkg, obj = resource_path.rsplit('.', 1)
    module = importlib.import_module(pkg)
    resource = getattr(module, obj)
    version = uuid4().hex
    target_module = importlib.import_module(package)
    target_dir = os.path.dirname(target_module.__file__)
    target_file = os.path.join(target_dir, resource.resource_type + '_' + version + '.py')

    with open(target_file, 'w') as f:
        f.write(MIGRATION_TEMPLATE.format(
            resource_path=resource_path,
            resource_type=resource.resource_type,
            version=version,
            previous_version=previous_version or '',
        ))

    return target_file


def collect(package='perch.migrations'):
    """
    Import all modules inside the perch.migrations package and return the
    registered migrations
    """
    package = importlib.import_module(package)

    for loader, name, is_pkg in pkgutil.walk_packages(package.__path__):
        importlib.import_module(package.__name__ + '.' + name)

    return _migrations


@coroutine
def run_migrations(migrations):
    """
    Run migrations for a resource type

    :param: a dicitionary of migrations
    """
    for resource, resource_migrations in migrations.items():
        for version in resource_migrations:
            to_migrate = yield resource_version.get(
                key=[resource.resource_type, version],
                include_docs=True)

            for x in to_migrate['rows']:
                instance = resource(**x['doc'])
                instance = _migrate_resource(
                    instance,
                    resource_migrations,
                    version
                )

                yield instance._save()


def _migrate_resource(instance, migrations, version=''):
    """
    Migrate a resource instance

    Subresources are migrated first, then the resource is recursively migrated

    :param instance: a perch.Document instance
    :param migrations: the migrations for a resource
    :param version: the current resource version to migrate
    """
    if version not in migrations:
        return instance

    instance = _migrate_subresources(
        instance,
        migrations[version]['subresources']
    )

    for migration in migrations[version]['migrations']:
        instance = migration(instance)
        instance._resource['doc_version'] = unicode(migration.version)

        instance = _migrate_resource(
            instance,
            migrations,
            version=migration.version
        )

    return instance


def _migrate_subresources(parent, migrations):
    """
    Migrate a resource's subresources

    :param parent: the parent perch.Document instance
    :param migrations: the migrations for a resource
    """
    for subresource, resource_migrations in migrations.items():
        parent = _migrate_subresource(
            subresource,
            parent,
            resource_migrations
        )

    return parent


def _migrate_subresource(subresource, parent, migrations):
    """
    Migrate a resource's subresource

    :param subresource: the perch.SubResource instance
    :param parent: the parent perch.Document instance
    :param migrations: the migrations for a resource
    """
    for key, doc in getattr(parent, subresource.parent_key, {}).items():
        for migration in migrations['migrations']:
            instance = migration(subresource(id=key, **doc))
            parent._resource['doc_version'] = unicode(migration.version)

            instance = _migrate_subresources(
                instance,
                migrations['subresources']
            )

        doc = instance._resource
        doc.pop('id', None)
        doc.pop(instance.resource_type + '_id', None)

        getattr(parent, subresource.parent_key)[key] = doc

    return parent



# -*- coding: utf-8 -*-
# Copyright 2016 Open Permissions Platform Coalition
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
from __future__ import unicode_literals

import uuid
from datetime import datetime, timedelta
from enum import Enum

import couch
from dateutil import parser
from passlib.hash import pbkdf2_sha256
from tornado.gen import coroutine, Return
from tornado.options import options, define
from voluptuous import All, Extra, In, Length, Required, Schema

from . import views, exceptions, validators
from .model import Document, SubResource, State


__all__ = ['User', 'UserOrganisation', 'Token']

define('min_length_password', default=3)
define('max_length_password', default=128)
define('token_ttl', default=86400)


class User(Document):
    resource_type = 'user'
    view = views.users
    active_view = views.active_users
    db_name = 'registry'
    internal_fields = (Document.internal_fields +
                       ['password', 'verification_hash'])
    read_only_fields = ['created_by']

    default_state = State.approved
    editable_states = [State.approved]

    # State transitions for users overridden so that:
    # - Only valid states are approved and deactivated
    # - Once user is deactivated cannot be reactivated
    approval_state_transitions = {}

    state_transitions = {
        None: [State.approved.name],
        State.approved.name: [State.deactivated.name]
    }

    class roles(Enum):
        administrator = 'administrator'
        user = 'user'
        default = user

    @property
    def schema(cls):
        orgs_schema = Schema({
            Extra: {
                'state': validators.validate_state,
                'role': In([x.value for x in cls.roles])
            }
        }, required=True)

        password_length = Length(min=options.min_length_password,
                                 max=options.max_length_password)

        schema = Schema({
            '_id': unicode,
            '_rev': unicode,
            'doc_version': unicode,
            Required('type', default=cls.resource_type): cls.resource_type,
            Required('email'): validators.valid_email,
            Required('password'): All(unicode, password_length),
            Required('has_agreed_to_terms'): True,
            Required('organisations', default={}): orgs_schema,
            Required('state', default=User.default_state.name): validators.validate_user_state,
            Required('role', default=User.roles.default.value): In([x.value for x in User.roles]),
            'first_name': unicode,
            'last_name': unicode,
            'phone': unicode,
            'verification_hash': unicode,
            'actor': unicode  # only used in fixtures, should it be removed?
        })

        return schema

    def clean(self):
        """Verified value is derived from whether user has a verification hash"""
        result = super(User, self).clean()
        result['verified'] = 'verification_hash' not in self._resource
        return result

    @coroutine
    def can_update(self, user, **kwargs):
        # Sys admin can update everything
        if user.is_admin():
            raise Return((True, set([])))

        # If user being updated can update everything apart from role
        if user.id == self.id:
            fields = ({'role'} & set(kwargs.keys()))
            if fields:
                raise Return((False, fields))
            else:
                raise Return((True, set([])))

        # Otherwise, cannot update
        raise Return((False, set([])))

    @coroutine
    def check_unique(self):
        """Check the user's email is unique"""
        emails = yield self.view.values(key=self.email)
        user_id = getattr(self, 'id', None)
        users = {x for x in emails if x != user_id and x}

        if users:
            raise exceptions.ValidationError(
                "User with email '{}' already exists".format(self.email))

    @classmethod
    @coroutine
    def create(cls, user, password, **kwargs):
        kwargs['verification_hash'] = unicode(uuid.uuid4().hex)

        resource = cls(password=cls.hash_password(password), **kwargs)
        yield resource._save()

        raise Return(resource)

    @classmethod
    @coroutine
    def create_admin(cls, email, password, **kwargs):
        """
        Create an approved 'global' administrator

        :param email: the user's email address
        :param password: the user's plain text password
        :returns: a User
        """
        data = {
            'email': email,
            'password': cls.hash_password(password),
            'has_agreed_to_terms': True,
            'state': State.approved,
            'role': cls.roles.administrator.value,
            'organisations': {}
        }
        data.update(**kwargs)

        user = cls(**data)
        yield user._save()

        raise Return(user)

    @staticmethod
    def hash_password(plain_text):
        """Hash a plain text password"""
        # NOTE: despite the name this is a one-way hash not a reversible cypher
        hashed = pbkdf2_sha256.encrypt(plain_text, rounds=8000, salt_size=10)
        return unicode(hashed)

    def verify_password(self, password):
        """Verify the password matches the hash"""
        return pbkdf2_sha256.verify(password, self.password)

    @coroutine
    def change_password(self, previous, new_password):
        """
        Change the user's password and save to the database

        :param previous: plain text previous password
        :param new_password: plain text new password
        :raises: ValidationError
        """
        if not self.verify_password(previous):
            raise exceptions.Unauthorized('Incorrect password')

        if len(new_password) < options.min_length_password:
            msg = ('Passwords must be at least {} characters'
                   .format(options.min_length_password))
            raise exceptions.ValidationError(msg)

        if len(new_password) > options.max_length_password:
            msg = ('Passwords must be at no more than {} characters'
                   .format(options.max_length_password))
            raise exceptions.ValidationError(msg)

        self.password = self.hash_password(new_password)
        yield self._save()

    @classmethod
    @coroutine
    def login(cls, email, password):
        """
        Log in a user

        :param email: the user's email address
        :param password: the user's password
        :returns: (User, token)
        :raises: SocketError, CouchException
        """
        try:
            doc = yield cls.view.first(key=email, include_docs=True)
        except couch.NotFound:
            raise exceptions.Unauthorized('Unknown email address')

        user = cls(**doc['doc'])

        verified = user.verify_password(password)
        if not verified:
            raise exceptions.Unauthorized('Invalid password')

        token = yield Token.create(user)
        raise Return((user, token.id))

    @classmethod
    @coroutine
    def verify(cls, user_id, verification_hash):
        """
        Verify a user using the verification hash

        The verification hash is removed from the user once verified

        :param user_id: the user ID
        :param verification_hash: the verification hash
        :returns: a User instance
        """
        user = yield cls.get(user_id)

        # If user does not have verification hash then this means they have already been verified
        if 'verification_hash' not in user._resource:
            raise Return(user)

        if user.verification_hash != verification_hash:
            raise exceptions.ValidationError('Invalid verification hash')

        del user.verification_hash
        yield user._save()

        raise Return(user)

    def is_admin(self):
        """Is the user a system administrator"""
        return self.role == self.roles.administrator.value and self.state == State.approved

    def is_org_admin(self, organisation_id):
        """Is the user authorized to administrate the organisation"""
        return (self._has_role(organisation_id, self.roles.administrator) or
                self.is_admin())

    def is_user(self, organisation_id):
        """Is the user authorized to administrate the organisation"""
        return (self._has_role(organisation_id, self.roles.user) or
                self.is_org_admin(organisation_id))

    def _has_role(self, organisation_id, role):
        """Check the user's role for the organisation"""
        if organisation_id is None:
            return False

        try:
            org = self.organisations.get(organisation_id, {})
            user_role = org.get('role')
            state = org.get('state')
        except AttributeError:
            return False

        return user_role == role.value and state == State.approved.name

    @classmethod
    @coroutine
    def remove_organisation_from_all(cls, organisation_id):
        """Remove an organisation from all users"""
        users = yield views.organisation_members.get(key=organisation_id,
                                                     include_docs=True)
        users = [x['doc'] for x in users['rows']]
        for user in users:
            user['organisations'][organisation_id]['state'] = State.deactivated.name

        db = cls.db_client()
        yield db.save_docs(users)


class Token(Document):
    resource_type = 'token'
    db_name = 'token'
    read_only_fields = ['token_ttl', 'user_id']

    @classmethod
    @coroutine
    def create(cls, user):
        token = cls(user_id=user.id)
        if 'token_ttl' not in token._resource:
            ttl = (datetime.utcnow() + timedelta(seconds=options.token_ttl))

        token.token_ttl = ttl.isoformat()
        yield token._save()

        raise Return(token)

    @property
    def ttl(self):
        return parser.parse(self.token_ttl)

    @classmethod
    @coroutine
    def valid(cls, token, **kwargs):
        """
        Check if a token exists and has not expired

        :param token: the token
        :return: bool
        """
        try:
            token = yield cls.get(token)
        except couch.NotFound:
            raise Return(False)

        raise Return(token.ttl >= datetime.utcnow())


class UserOrganisation(SubResource):
    resource_type = 'user_organisation'
    parent_resource = User
    parent_key = 'organisations'
    view = views.user_organisations_resource
    active_view = views.active_user_organisations_resource
    internal_fields = ['id', 'user_id']
    editable_states = [State.approved]

    schema = Schema({
        'id': unicode,
        'user_id': unicode,
        Required('state', default=SubResource.default_state.name): validators.validate_state,
        Required('role', default=User.roles.default.value): In([x.value for x in User.roles])
    }, )

    @property
    def organisation_id(self):
        return self.id

    @coroutine
    def can_approve(self, user, **data):
        """
        Only org admins can approve joining an organisation
        :param user: a User
        :param data: data that the user wants to update
        """
        is_org_admin = user.is_org_admin(self.organisation_id)
        raise Return(is_org_admin)

    @coroutine
    def can_update(self, user, **kwargs):
        # Can only update if org admin or user being updated
        if not (user.id == self.user_id or user.is_org_admin(self.organisation_id)):
            raise Return((False, set([])))

        if 'role' in kwargs:
            if not user.is_org_admin(self.organisation_id):
                raise Return((False, {'role'}))

        raise Return((True, set([])))



# -*- coding: utf-8 -*-
# Copyright 2016 Open Permissions Platform Coalition
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
from __future__ import unicode_literals

import random
import string
import uuid
from copy import deepcopy
from collections import defaultdict
from operator import itemgetter

import couch
from tornado.gen import coroutine, Return
from tornado.options import options, define
from voluptuous import (
    All, Any, Extra, In, Invalid, Length,
    MultipleInvalid, Range, Required, Schema)

from . import views, exceptions, validators
from .model import format_error, Document, SubResource, State
from .validators import MetaSchema, partial_schema


__all__ = ['Organisation', 'Service', 'Repository', 'OAuthSecret']

define('min_length_organisation_name', default=3)
define('max_length_organisation_name', default=512)
define('min_length_service_name', default=3)
define('max_length_service_name', default=128)
define('min_length_repository_name', default=3)
define('max_length_repository_name', default=512)

SERVICE_TYPES = {
    'external',
    'identity',
    'index',
    'onboarding',
    'query',
    'repository',
    'resolution',
    'transformation',
}
PERMISSIONS = ['-', 'r', 'w', 'rw']


class Organisation(Document):
    resource_type = 'organisation'
    db_name = 'registry'
    internal_fields = Document.internal_fields + ['services', 'repositories']
    read_only_fields = ['created_by']

    @property
    def schema(self):
        name_length = Length(min=options.min_length_organisation_name,
                             max=options.max_length_organisation_name)

        filtered_fields = ['id', 'organisation_id']

        schema = Schema({
            Required('name'): All(unicode, name_length),
            Required('state', default=self.default_state.name): validators.validate_state,
            Required('created_by'): unicode,
            Required('type', default=self.resource_type): self.resource_type,
            Required('star_rating', default=0): Range(0, 5),
            Required('services', default={}): {
                Extra: partial_schema(Service.schema, filtered_fields)
            },
            Required('repositories', default={}): {
                Extra: partial_schema(Repository.schema, filtered_fields)
            },
            '_id': unicode,
            '_rev': unicode,
            'doc_version': unicode,
            'description': unicode,
            'address': unicode,
            'email': Any(validators.valid_email, ''),
            'phone': unicode,
            'website': unicode,
            'facebook': unicode,
            'twitter': unicode,
            'google_plus': unicode,
            'instagram': unicode,
            'youtube': unicode,
            'linkedin': unicode,
            'myspace': unicode,
            'payment': {'url': validators.validate_url},
            'reference_links': validators.validate_reference_links,
            'logo': Any(validators.validate_url, '', None),
            'primary_color': validators.validate_hex,
            'secondary_color': validators.validate_hex,
            'modal_header_text': unicode,
            'modal_footer_text': unicode,
            'modal_link_text': unicode,
            'modal_link_url': Any(validators.validate_url, '', None)
        })

        return schema

    @coroutine
    def create_default_service(self, user):
        service = yield Service.create(
            user,
            created_by=user.id,
            name=unicode(uuid.uuid4().hex),
            service_type='external',
            organisation_id=self.id
        )
        self.services[service.id] = service._resource

    @classmethod
    @coroutine
    def create(cls, user, **kwargs):
        if not user.is_admin():
            # Force star rating to be 0 if the user is not an admin
            kwargs['star_rating'] = 0

        organisation = yield super(Organisation, cls).create(user, **kwargs)

        # If organisation is approved on creation, create default service
        if organisation.state == State.approved:
            yield organisation.create_default_service(user)

        raise Return(organisation)

    @coroutine
    def check_unique(self):
        result = yield views.organisation_name.values(key=self.name)

        org_id = getattr(self, 'id', None)
        orgs = {x for x in result if x != org_id and x}

        if orgs:
            raise exceptions.ValidationError(
                "Organisation with name '{}' already exists".format(self.name))

    @classmethod
    @coroutine
    def all(cls, state=None, include_deactivated=False):
        """
        Get all organisations

        :param state: State of organisation
        :param include_deactivated: Flag to include deactivated
        :returns: list of Organisation instances
        :raises: SocketError, CouchException
        """
        if state and state not in validators.VALID_STATES:
            raise exceptions.ValidationError('Invalid "state"')
        elif state:
            organisations = yield views.organisations.get(key=state,
                                                          include_docs=True)
        elif include_deactivated:
            organisations = yield views.organisations.get(include_docs=True)
        else:
            organisations = yield views.active_organisations.get(include_docs=True)

        raise Return([cls(**org['doc']) for org in organisations['rows']])

    @classmethod
    @coroutine
    def user_organisations(cls, user_id, state=None, include_deactivated=False):
        """
        Get organisations that the user has joined

        :param user_id: the user ID
        :param state: the user's "join" state
        :param include_deactivated: Include deactivated resources in response
        :returns: list of Organisation instances
        :raises: SocketError, CouchException
        """
        if state and state not in validators.VALID_STATES:
            raise exceptions.ValidationError('Invalid "state"')

        if include_deactivated:
            organisations = yield views.joined_organisations.get(
                key=[user_id, state], include_docs=True)
        else:
            organisations = yield views.active_joined_organisations.get(
                key=[user_id, state], include_docs=True)

        raise Return([cls(**org['doc']) for org in organisations['rows']])

    @coroutine
    def update(self, user, **kwargs):
        previous_state = self.state

        yield super(Organisation, self).update(user, **kwargs)

        approved = self.state == State.approved
        was_pending = previous_state == State.pending
        if approved and was_pending:
            yield self.create_default_service(user)

    @coroutine
    def can_update(self, user, **data):
        """
        Sys admins can always update an organisation.

        Organisation admins and creators can update, but may not update the following fields:

            - star_rating

        :param user: a User
        :param data: data that the user wants to update
        :returns: bool, set of fields that the user was not authorized to update
        """
        if user.is_admin():
            raise Return((True, set([])))

        org_admin = user.is_org_admin(self.id)
        creator = self.created_by == user.id
        if org_admin or creator:
            fields = {'star_rating'} & set(data.keys())
            if fields:
                raise Return((False, fields))
            else:
                raise Return((True, set([])))

        raise Return((False, set([])))

    @coroutine
    def can_approve(self, user, **data):
        """
        Only sys admins can approve an organisation
        :param user: a User
        :param data: data that the user wants to update
        """
        is_admin = user.is_admin()
        raise Return(is_admin)


all_permission_schema = Schema({
    'type': 'all',
    'permission': In(PERMISSIONS),
    'value': None
}, required=True)
organisation_permission_schema = all_permission_schema.extend({
    'type': 'organisation_id',
    'permission': In(PERMISSIONS),
    'value': unicode
})
service_type_permission_schema = all_permission_schema.extend({
    'type': 'service_type',
    'permission': In(PERMISSIONS),
    'value': In(SERVICE_TYPES)
})


def group_permissions(permissions):
    """
    Groups a permissions list

    Returns a dictionary, with permission types as keys and sets of entities
    with access to the resource as values, e.g.:

        {
            'organisation_id': {
                'org1': set(['rw', 'r', 'w']),
                'org2': set(['-']),
                'org3': set(['r', 'w']),
            },
            'all': set(['r'])
        }

    'org1' has 'rw' access to the resource, 'org2' is denied access and 'org3'
    has 'r' & 'w' access (the same as 'rw'). Note that 'rw' will always result
    in 'rw', 'r' & 'w' in the set to make checks easier.

    If present in the resource's permissions, the 'all' permission type is an
    exception in that it's value is just a set instead of a dictionary.

    :param permissions: a list of permissions
    :returns: defaultdict
    """
    groups = defaultdict(lambda: defaultdict(set))

    for p in sorted(permissions, key=itemgetter('type')):
        permission_set = groups[p['type']][p.get('value')]
        permission_set.add(p['permission'])

        if p['permission'] == 'rw':
            permission_set.update({'r', 'w'})

    # the 'all' permission type always has None as the value
    groups['all'] = groups['all'][None]

    return groups

service_name_length = Length(min=options.min_length_service_name,
                             max=options.max_length_service_name)


def validate_service_schema(v):
    if v['service_type'] != 'external' and 'location' not in v:
        raise Invalid('location is required')
    return v


class Service(SubResource):
    resource_type = 'service'
    parent_resource = Organisation
    parent_key = 'services'
    read_only_fields = ['created_by']
    view = views.services
    active_view = views.active_services

    default_permission = [{'type': 'all', 'value': None, 'permission': 'rw'}]
    schema = MetaSchema({
        Required('id', default=lambda: unicode(uuid.uuid4().hex)): unicode,
        Required('name'): All(unicode, service_name_length),
        Required('organisation_id'): unicode,
        Required('permissions', default=default_permission): [Any(
            all_permission_schema,
            organisation_permission_schema,
            service_type_permission_schema
        )],
        Required('type', default=resource_type): resource_type,
        Required('created_by'): unicode,
        Required('service_type'): In(SERVICE_TYPES),
        Required('state', default=SubResource.default_state.name): validators.validate_state,
        'location': validators.validate_url
    }, validate_service_schema)

    @coroutine
    def check_unique(self):
        """Check the service's name and location are unique"""
        errors = []
        service_id = getattr(self, 'id', None)
        fields = [('location', views.service_location),
                  ('name', views.service_name)]

        for field, view in fields:
            value = getattr(self, field, None)
            if not value:
                continue

            result = yield view.values(key=value)
            matched = {x['id'] for x in result if x['id'] != service_id}
            if matched:
                errors.append("Service with {} '{}' already exists"
                              .format(field, value))

        if errors:
            raise exceptions.ValidationError(errors)

    def clean(self, user=None):
        """Remove internal fields"""
        doc = self._resource
        internal_fields = deepcopy(self.internal_fields)
        if user is None or not user.is_user(self.organisation_id):
            internal_fields.append('permissions')

        result = {k: v for k, v in doc.iteritems() if k not in internal_fields}

        return result

    @classmethod
    @coroutine
    def get_by_location(cls, location, include_deactivated=False):
        """Get a service by it's location"""
        if include_deactivated:
            view = views.service_location
        else:
            view = views.active_service_location

        result = yield view.first(key=location, include_docs=True)
        parent = cls.parent_resource(**result['doc'])

        raise Return(cls(parent=parent, **result['value']))

    @classmethod
    @coroutine
    def all(cls, service_type=None, organisation_id=None, include_deactivated=False):
        """
        Get all resources

        :param service_type: Filter by service type
        :param organisation_id: Filter by organisation id
        :param include_deactivated: Flag to include deactivated Services

        :returns: list of Resource instances
        :raises: SocketError, CouchException
        """
        if include_deactivated:
            resources = yield views.services.get(key=[service_type, organisation_id])
        else:
            resources = yield views.active_services.get(key=[service_type, organisation_id])

        # TODO: shouldn't this include the doc as the parent?
        raise Return([cls(**resource['value']) for resource in resources['rows']])

    @classmethod
    @coroutine
    def create(cls, user, **kwargs):
        resource = yield super(Service, cls).create(user, **kwargs)

        # If service is approved on creation, create secret for service
        if resource.state == State.approved:
            yield OAuthSecret.create(user, client_id=resource.id)

        raise Return(resource)

    @classmethod
    @coroutine
    def can_create(cls, user, **kwargs):
        raise Return(user.is_user(kwargs['organisation_id']))

    @coroutine
    def update(self, user, **kwargs):
        previous_state = self.state
        yield super(Service, self).update(user, **kwargs)

        approved = self.state == State.approved
        was_pending = previous_state == State.pending
        if approved and was_pending:
            secrets = yield OAuthSecret.view.get(key=self.id)
            if not secrets['rows']:
                yield OAuthSecret.create(user, client_id=self.id)

    @coroutine
    def can_approve(self, user, **data):
        """
        Only sys admins can approve a service
        :param user: a User
        :param data: data that the user wants to update
        """
        is_external = data.get('service_type', self.service_type) == 'external'
        raise Return(user.is_admin() or is_external)

    @coroutine
    def can_update(self, user, **kwargs):
        """Org admins may not update organisation_id or service_type"""
        if user.is_admin():
            raise Return((True, set([])))

        is_creator = self.created_by == user.id
        if not (user.is_org_admin(self.organisation_id) or is_creator):
            raise Return((False, set([])))

        fields = ({'service_type', 'organisation_id'} & set(kwargs.keys()))
        if fields:
            raise Return((False, fields))
        else:
            raise Return((True, set([])))

    @classmethod
    @coroutine
    def authenticate(cls, client_id, secret):
        """
        Authenticate a client using it's secret

        :param client_id: the client / service ID
        :param secret: the client secret
        :returns: a Service instance
        """
        result = yield views.oauth_client.get(key=[secret, client_id])
        if not result['rows']:
            raise Return()

        service = yield Service.get(client_id)
        raise Return(service)

    def authorized(self, requested_access, resource):
        """
        Check whether the service is authorized to access the resource

        :param requested_access: "r", "w", or "rw"
        :param resource: a Resource or SubResource with "permissions" attribute
        :returns: True if has access, False otherwise
        """
        if {self.state, resource.state} != {State.approved}:
            return False

        permissions = group_permissions(getattr(resource, 'permissions', []))

        org_permissions = permissions['organisation_id'][self.organisation_id]
        type_permissions = permissions['service_type'][self.service_type]
        all_permissions = permissions['all']

        for permission_set in [org_permissions, type_permissions, all_permissions]:
            if '-' in permission_set:
                return False
            elif set([x for x in requested_access]).issubset(permission_set):
                return True

        return False


class Repository(SubResource):
    resource_type = 'repository'
    parent_resource = Organisation
    parent_key = 'repositories'
    read_only_fields = ['created_by']
    view = views.repositories
    active_view = views.active_repositories

    # State transitions for repositories overridden so that:
    # - Cannot deactivate an approved repository
    # - Repository can move from approved to pending when moved to a new repository service
    approval_state_transitions = {
        None: [State.approved.name],
        State.pending.name: [State.approved.name, State.rejected.name]
    }

    state_transitions = {
        None: [State.pending.name],
        State.pending.name: [State.deactivated.name],
        State.approved.name: [State.pending.name],
        State.rejected.name: [State.deactivated.name],
        State.deactivated.name: [State.pending.name]
    }

    _repository_name_length = Length(min=options.min_length_repository_name,
                                     max=options.max_length_repository_name)

    schema = Schema({
        Required('id', default=lambda: unicode(uuid.uuid4().hex)): unicode,
        Required('name'): All(unicode, _repository_name_length),
        Required('service_id'): unicode,
        Required('organisation_id'): unicode,
        Required('state', default=SubResource.default_state.name): validators.validate_state,
        Required('type', default=resource_type): resource_type,
        Required('created_by'): unicode,
        Required('permissions'): [Any(
            all_permission_schema,
            organisation_permission_schema
        )],
    })

    def clean(self, user=None):
        """Remove internal fields"""
        doc = self._resource
        internal_fields = deepcopy(self.internal_fields)
        if user is None or not user.is_user(self.organisation_id):
            internal_fields.append('permissions')

        result = {k: v for k, v in doc.iteritems() if k not in internal_fields}

        return result

    @property
    def default_permissions(self):
        default_permissions = [
            {'type': 'all', 'value': None, 'permission': 'r'},
        ]

        if 'organisation_id' in self._resource:
            default_permissions.append({
                'type': 'organisation_id',
                'value': self.organisation_id,
                'permission': 'rw'
            })

        return default_permissions

    @coroutine
    def validate(self):
        """Validate the resource"""
        if not self._resource.get('permissions'):
            self.permissions = self.default_permissions

        try:
            # update _resource so have default values from the schema
            self._resource = self.schema(self._resource)
        except MultipleInvalid as e:
            errors = [format_error(err, self.resource_type) for err in e.errors]
            raise exceptions.ValidationError({'errors': errors})

        yield self.check_service()
        yield self.check_unique()

    @coroutine
    def check_service(self):
        """Check the service exists and is a repository service"""
        try:
            service = yield Service.get(self.service_id)
        except couch.NotFound:
            raise exceptions.ValidationError('Service {} not found'
                                             .format(self.service_id))

        if service.service_type != 'repository':
            raise exceptions.ValidationError('{} is not a repository service'
                                             .format(self.service_id))

        if service.state != State.approved:
            raise exceptions.ValidationError('{} is not an approved service'
                                             .format(self.service_id))

    @coroutine
    def check_unique(self):
        """Check the repository's name is unique"""
        result = yield views.repository_name.values(key=self.name)
        repo_id = getattr(self, 'id', None)
        repos = {x for x in result if x != repo_id and x}

        if repos:
            raise exceptions.ValidationError(
                "Repository with name '{}' already exists".format(self.name))

    @coroutine
    def update(self, user, **kwargs):
        service_id = kwargs.get('service_id')
        if service_id:
            can_approve = yield self.can_approve(user, **kwargs)
            if not can_approve:
                kwargs['state'] = State.pending.name

        yield super(Repository, self).update(user, **kwargs)

    @coroutine
    def can_approve(self, user, **data):
        """
        Admins of repository service or sys admins can approve a repository
        :param user: a User
        :param data: data that the user wants to update
        """
        service_id = data.get('service_id', self.service_id)

        try:
            service = yield Service.get(service_id)

            is_repo_admin = user.is_org_admin(service.organisation_id)
            raise Return(is_repo_admin)
        except couch.NotFound:
            pass

        raise Return(False)

    @coroutine
    def can_update(self, user, **kwargs):
        """
        Sys admin's can change anything

        If the user is an organisation administrator or created the repository,
        they may change any field other than "organisation_id"

        If the user is a service administrator the user may change the "state"
        but no other fields.
        """
        if user.is_admin():
            raise Return((True, set([])))

        is_creator = self.created_by == user.id
        if user.is_org_admin(self.organisation_id) or is_creator:
            fields = set([])
            if 'organisation_id' in kwargs:
                fields.add('organisation_id')

            if fields:
                raise Return((False, fields))
            else:
                raise Return((True, set([])))

        try:
            service = yield Service.get(self.service_id)

            if user.is_org_admin(service.organisation_id):
                fields = set(kwargs) - {'state'}
                if fields:
                    raise Return((False, fields))
                else:
                    raise Return((True, fields))
        except couch.NotFound:
            # will be handled in Repository.validate
            pass

        raise Return((False, set([])))

    @classmethod
    @coroutine
    def can_create(cls, user, **kwargs):
        return user.is_user(kwargs.get('organisation_id'))

    @coroutine
    def with_relations(self, user=None):
        """
        Return a cleaned dictionary including relations

        :returns: a Repository instance
        """
        repository = self.clean(user=user)
        try:
            parent = yield self.get_parent()
            repository['organisation'] = parent.clean()
        except couch.NotFound:
            parent = None
            repository['organisation'] = {'id': self.parent_id}

        service_id = self.service_id
        try:
            # TODO: cache this lookup
            service = yield Service.get(service_id)
            repository['service'] = service.clean(user=user)
        except couch.NotFound:
            # just include the service ID if cannot find the service
            repository['service'] = {'id': service_id}

        del repository['service_id']
        del repository['organisation_id']

        raise Return(repository)


def generate_secret(length=30):
    """
    Generate an ASCII secret using random.SysRandom

    Based on oauthlib's common.generate_token function
    """
    rand = random.SystemRandom()
    ascii_characters = string.ascii_letters + string.digits

    return ''.join(rand.choice(ascii_characters) for _ in range(length))


class OAuthSecret(Document):
    db_name = 'registry'
    resource_type = 'oauth_client_credentials'
    view = active_view = views.oauth_secrets

    schema = Schema({
        Required('_id', default=generate_secret): unicode,
        Required('client_id'): unicode,
        Required('type', default=resource_type): resource_type,
        '_rev': unicode,
        'doc_version': unicode
    })

    @classmethod
    @coroutine
    def can_create(cls, user, **kwargs):
        service = yield Service.get(kwargs['client_id'])
        creator = service.created_by == user.id
        can_create = user.is_org_admin(service.organisation_id) or creator
        approved = service.state == State.approved

        raise Return(can_create and approved)

    @coroutine
    def can_delete(self, user):
        service = yield Service.get(self.client_id)
        creator = service.created_by == user.id
        raise Return(user.is_org_admin(service.organisation_id) or creator)

    @property
    def secret(self):
        return self._id

    @classmethod
    @coroutine
    def client_secrets(cls, client_id):
        """
        Get the client's secrets using the client_id

        :param client_id: the client ID, e.g. a service ID
        :returns: list OAuthSecret instances
        """
        secrets = yield cls.view.get(key=client_id, include_docs=True)
        raise Return([cls(**secret['doc']) for secret in secrets['rows']])

    @classmethod
    @coroutine
    def delete_all_secrets(cls, user, client_id):
        """Delete all of the client's credentials"""
        can_delete = yield cls(client_id=client_id).can_delete(user)
        if not can_delete:
            raise exceptions.Unauthorized('User may not delete {} secrets'
                                          .format(client_id))

        results = yield cls.view.get(key=client_id, include_docs=True)

        if results['rows']:
            db = cls.db_client()

            docs = [{
                '_rev': doc['doc']['_rev'],
                '_id': doc['doc']['_id'],
                '_deleted': True
            } for doc in results['rows']]

            yield db.save_docs(docs)



# -*- coding: utf-8 -*-
# Copyright 2016 Open Permissions Platform Coalition
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.

from tornado.options import define
from .model import *
from .user import *
from .organisation import *
from . import migrate

__version__ = '0.15.11'

define('db_port', default=5984)
define('url_registry_db', default='http://localhost')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2016 Open Permissions Platform Coalition
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations under
# the License.
import logging
from functools import partial

import click
from tornado.ioloop import IOLoop
from tornado.options import options

from . import migrate


@click.group()
def cli():
    pass


@click.group()
def migrations():
    pass


@click.command()
@click.option('--migrations_path', help='Python path to migrations package, e.g. perch.migrations')
@click.option('--previous', help='The previous version of the resource to migrate')
@click.argument('resource')
def create(resource, previous=None, migrations_path=None):
    """Create an empty migration for a resource"""
    if migrations_path:
        file_path = migrate.create(resource, previous_version=previous, package=migrations_path)
    else:
        file_path = migrate.create(resource, previous_version=previous)

    click.secho('Created migration file: ' + file_path, fg='green')

@click.command()
@click.option('--port', help='The database port')
@click.option('--url', help='The database URL')
@click.option('--migrations_path', help='Python path to migrations package, e.g. perch.migrations')
def run(migrations_path=None, url=None, port=None):
    """Run migrations"""
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)

    if url:
        url = str(url).rstrip('/')
        options.url_registry_db = url

    if port:
        options.db_port = int(port)

    if migrations_path:
        migrations = migrate.collect(migrations_path)
    else:
        migrations = migrate.collect()

    func = partial(migrate.run_migrations, migrations)
    IOLoop.instance().run_sync(func)

migrations.add_command(create)
migrations.add_command(run)

cli.add_command(migrations)


if __name__ == '__main__':
    cli()



# -*- coding: utf-8 -*-
# Copyright 2016 Open Permissions Platform Coalition
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.

"""Voluptuous validotor functions"""
import re
from urlparse import urlsplit

from voluptuous import AllInvalid, Invalid, Schema, ALLOW_EXTRA

from .model import State


class MetaSchema(object):
    """
    Schema must pass all validators. Useful for cases where a field depends on
    the value of another field

    Similar to using All with a schema and vaildator function, e.g.
    All(Schema({'x': int, 'y': int}), x_greater_than_y)

    >>> validate = MetaSchema({'x': '10'}, Coerce(int))
    >>> validate('10')
    10
    """

    def __init__(self, schema, *validators, **kwargs):
        self.validators = validators
        self.msg = kwargs.pop('msg', None)
        self._schema = Schema(schema)
        self._schemas = [Schema(val, **kwargs) for val in validators]

    @property
    def schema(self):
        return self._schema.schema

    def __call__(self, v):
        try:
            v = self._schema(v)
            for schema in self._schemas:
                v = schema(v)
        except Invalid as e:
            raise e if self.msg is None else AllInvalid(self.msg)

        return v


def partial_schema(schema, filtered_fields):
    """
    Validator for part of a schema, ignoring some fields

    :param schema: the Schema
    :param filtered_fields: fields to filter out
    """
    return Schema({
        k: v for k, v in schema.schema.items()
        if getattr(k, 'schema', k) not in filtered_fields
    }, extra=ALLOW_EXTRA)


def valid_email(email):
    """Validate email."""
    if "@" not in email:
        raise Invalid('This email is invalid.')
    return email


def validate_hex(color):
    """
    Validate string is a hex color code
    """
    hex_re = '^#(?:[0-9a-fA-F]{3}){1,2}$'
    if not re.match(hex_re, color):
        raise Invalid('Invalid Hex Color')
    return color


def validate_url(url):
    """Validate URL is valid

    NOTE: only support http & https
    """
    schemes = ['http', 'https']
    netloc_re = re.compile(
        r'^'
        r'(?:\S+(?::\S*)?@)?'  # user:pass auth
        r'(?:[a-z0-9]|[a-z0-9][a-z0-9\-]{0,61}[a-z0-9])'
        r'(?:\.(?:[a-z0-9]|[a-z0-9][a-z0-9\-]{0,61}[a-z0-9]))*'  # host
        r'(?::[0-9]{2,5})?'  # port
        r'$', re.IGNORECASE
    )

    try:
        scheme, netloc, path, query, fragment = urlsplit(url)
    except ValueError:
        raise Invalid('Invalid URL')

    if scheme not in schemes:
        raise Invalid('Missing URL scheme')

    if not netloc_re.search(netloc):
        raise Invalid('Invalid URL')

    return url


def validate_reference_links(reference_links):
    """
    Vaidate reference links data structure

    Expected data structure:
        {
            "links": {
                id_type1: url1,
                id_type2: url2
            },
            "redirect_id_type": id_type1 | id1_type2
        }

    where links is an optional key but must be a dictionary with id types to
    URLs if it exists, and redirect_id_type is optional but if it exists,
    it must point to one of the existing id types in the links object. It is
    used to set a default redirect URL that is used by the resolution service.
    """
    allowed_keys = ['links', 'redirect_id_type']

    if not isinstance(reference_links, dict):
        raise Invalid('Expected reference_links to be an object')

    if 'links' in reference_links and not isinstance(reference_links['links'], dict):
        raise Invalid('Expected links in reference_links to be an object')

    links = reference_links.get('links', {})
    redirect_id_type = reference_links.get('redirect_id_type')

    for key in reference_links:
        if key not in allowed_keys:
            raise Invalid('Key {} is not allowed'.format(key))

    if redirect_id_type and redirect_id_type not in links:
        raise Invalid('Redirect ID type must point to one of the links\' ID types')

    [validate_url(url) for url in links.values()]

    return reference_links


VALID_STATES = {x.name for x in State}
VALID_USER_STATES = {x.name for x in [State.approved, State.deactivated]}


def validate_state(state):
    return _validate_state(state, VALID_STATES)


def validate_user_state(state):
    return _validate_state(state, VALID_USER_STATES)


def _validate_state(state, valid_states):
    """Validate a state string"""
    if state in State:
        return state.name
    elif state in valid_states:
        return state
    else:
        raise Invalid('Invalid state')



# -*- coding: utf-8 -*-
# Copyright 2016 Open Permissions Platform Coalition
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.

from __future__ import unicode_literals

from copy import deepcopy
from enum import Enum
from operator import attrgetter

import couch
from tornado.gen import coroutine, Return
from tornado.options import options
from voluptuous import MultipleInvalid, Schema, ALLOW_EXTRA, Required, Undefined

from . import exceptions


__all__ = ['State', 'Document', 'SubResource']


class State(Enum):
    """
    Resource states, the enum values represent the priority of the state.

    For example, rejected has higher priority than approved
    """
    approved = 0
    pending = 1
    rejected = 2
    deactivated = 3


def format_error(invalid, doc_type):
    """
    format the error message using the voluptuous.Invalid object
    :param invalid: voluptuous.Invalid instance
    :param doc_type: type of the object, e.g. Organisation, User
    """
    # using string for checking is probably not ideal,
    # but voluptuous does not have specific sub error
    # types for these errors
    if invalid.error_message == 'extra keys not allowed':
        msg = "Key '{}' is not allowed".format(invalid.path[0])
    elif invalid.error_message == 'required key not provided':
        msg = "{} '{}' is missing".format(doc_type, invalid.path[0])
    else:
        msg = invalid.message
    return {'message': msg, 'field': str(invalid.path[0])}


class Document(object):
    internal_fields = ['type', '_id', '_rev', 'doc_version']
    _resource = {}
    resource_type = None
    schema = Schema({
        '_id': unicode,
        '_rev': unicode,
        'doc_version': unicode
    }, extra=ALLOW_EXTRA)
    # read only fields can not be changed after the resource has been created
    read_only_fields = []

    default_state = State.pending
    editable_states = [State.approved, State.pending]

    # State transitions that can be performed by users with permission to update resource
    state_transitions = {
        None: [State.pending.name],
        State.pending.name: [State.deactivated.name],
        State.approved.name: [State.deactivated.name],
        State.rejected.name: [State.deactivated.name],
        State.deactivated.name: [State.pending.name]
    }

    # State transitions that can be performed by users with permission to approve resource
    # These are in addition to regular state_transitions
    approval_state_transitions = {
        None: [State.approved.name],
        State.pending.name: [State.approved.name, State.rejected.name],
        State.deactivated.name: [State.approved.name],
    }

    def __init__(self, **kwargs):
        self._resource = deepcopy(kwargs)
        self._resource['type'] = self.resource_type

    def __getattr__(self, attr):
        try:
            if attr in self._resource:
                return self._resource[attr]
            else:
                defaults = self.get_required_fields_with_defaults()
                return defaults[attr]
        except KeyError:
            raise AttributeError(attr)

    def __setattr__(self, attr, value):
        if attr in dir(self):
            object.__setattr__(self, attr, value)
        elif attr in self._read_only and attr in self._resource:
            raise exceptions.ValidationError('{} may not be modified'
                                             .format(attr))
        else:
            self._resource[attr] = value

    def __delattr__(self, attr):
        if attr in dir(self):
            raise AttributeError('"{}" is cannot be deleted'.format(attr))
        elif attr in self._read_only and attr in self._resource:
            raise exceptions.ValidationError('{} may not be modified'
                                             .format(attr))

        try:
            del self._resource[attr]
        except KeyError:
            raise AttributeError(attr)

    @property
    def _read_only(self):
        return set(self.read_only_fields) | {'_id', '_rev', 'type', 'doc_version'}

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        self._id = value

    @property
    def editable(self):
        return self.state in self.editable_states

    @coroutine
    def validate(self):
        """Validate the resource using its voluptuous schema"""
        try:
            # update _resource to have default values from the schema
            self._resource = self.schema(self._resource)
        except MultipleInvalid as e:
            errors = [format_error(err, self.resource_type) for err in e.errors]
            raise exceptions.ValidationError({'errors': errors})

        yield self.check_unique()

    def get_required_fields_with_defaults(self):
        defaults = {}
        required_keys = set(
            key for key in self.schema.schema if isinstance(key, Required))
        for key in required_keys:
            if not isinstance(key.default, Undefined):
                defaults[unicode(key)] = key.default()
        return defaults

    @coroutine
    def check_unique(self):
        """
        Override this method to check the resource is unique

        If a unique field isn't unique raise a ValidationError
        """
        pass

    @classmethod
    def db_client(cls):
        db_url = ':'.join([options.url_registry_db, str(options.db_port)])
        return couch.AsyncCouch(db_name=cls.db_name, couch_url=db_url)

    @classmethod
    @coroutine
    def get(cls, resource_id, include_deactivated=False):
        """
        Get a resource

        :param resource_id: a resource's ID
        :param include_deactivated: Include deactivated resources in response

        :returns: Document instance
        :raises: SocketError, CouchException
        """
        resource = yield cls.db_client().get_doc(resource_id)

        if resource.get('type') != cls.resource_type:
            raise exceptions.NotFound()

        if not include_deactivated and resource.get('state') == State.deactivated.name:
            raise exceptions.NotFound()

        raise Return(cls(**resource))

    @classmethod
    @coroutine
    def all(cls, include_deactivated=False):
        """
        Get all resources
        :param include_deactivated: Include deactivated resources in response

        :returns: list of Document instances
        :raises: SocketError, CouchException
        """
        if include_deactivated:
            resources = yield cls.view.get(include_docs=True)
        else:
            resources = yield cls.active_view.get(include_docs=True)
        raise Return([cls(**resource['doc']) for resource in resources['rows']])

    @classmethod
    @coroutine
    def create(cls, user, **kwargs):
        resource = cls(**kwargs)

        if user:
            can_create = yield cls.can_create(user, **kwargs)
            can_approve = yield resource.can_approve(user, **kwargs)
        else:
            can_create = False
            can_approve = False

        if not can_create:
            err = 'User is not authorised to create this resource'
            raise exceptions.Unauthorized(err)

        # If user can approve resource and has not specified a state, approve on creation
        if 'state' not in kwargs and can_approve:
            resource._resource.update({'state': State.approved.name})

        can_set_state = yield resource.validate_state_transition(user, None, resource._resource.get('state'), **kwargs)
        if not can_set_state:
            err = [{
                'field': 'state',
                'message': 'Cannot set initial state as {}'.format(resource._resource.get('state'))
            }]
            raise exceptions.Unauthorized({'errors': err})

        yield resource._save()

        raise Return(resource)

    @coroutine
    def validate_state_transition(self, user, start_state, end_state, **kwargs):
        """
        Validate whether user can transition resource from start state to end state
        :param user
        :param start_state
        :param end_state
        :return: bool
        """
        if start_state == end_state:
            raise Return(True)

        transitions = self.state_transitions.get(start_state, [])

        approved_transitions = []
        can_approve = yield self.can_approve(user, **kwargs)
        if can_approve:
            approved_transitions = self.approval_state_transitions.get(start_state, [])

        if end_state not in transitions and end_state not in approved_transitions:
            raise Return(False)

        raise Return(True)

    @coroutine
    def update(self, user, **kwargs):
        if user:
            can_update, fields = yield self.can_update(user, **kwargs)
        else:
            can_update = False
            fields = []

        new_state = getattr(State, kwargs.get('state', ''), self.state)
        can_edit = {self.state, new_state}.intersection(set(self.editable_states))
        if not can_edit:
            err = 'User is not authorised to update resource with state {}'.format(new_state.name)
            raise exceptions.Unauthorized(err)

        if not can_update:
            if fields:
                msg = "User cannot update field '{}'"
                err = [{
                    'field': x,
                    'message': msg.format(x)
                } for x in fields]
                raise exceptions.Unauthorized({'errors': err})
            else:
                err = 'User is not authorised to update this resource'
                raise exceptions.Unauthorized(err)

        fields_to_update = set(kwargs.keys()) & set(self._resource.keys())

        state_field_update = fields_to_update & {'state'}

        if state_field_update:
            can_transition = yield self.validate_state_transition(user, self._resource.get('state'),
                                                                  kwargs['state'], **kwargs)
            if not can_transition:
                err = [{
                    'field': 'state',
                    'message': ('Cannot transition from {} state to {} state'
                                .format(self.state.name, kwargs['state']))
                }]
                raise exceptions.Unauthorized({'errors': err})

        read_only_fields = fields_to_update & self._read_only
        if read_only_fields:
            msg = '{} may not be modified'
            errors = [msg.format(x) for x in read_only_fields]
            raise exceptions.ValidationError(errors)

        self._resource.update(kwargs)

        yield self._save()

    @coroutine
    def _save(self):
        """
        Save the resource

        It's better to use the create, updated & delete methods intsead of
        modifying an instance and calling save, because then we call the
        can_create, can_update & can_delete methods to check whether a user
        is permitted to make the changes.
        """
        yield self.validate()

        db = self.db_client()
        saved = yield db.save_doc(self._resource)

        # Allow couch to create Document IDs
        if '_id' not in self._resource:
            self._resource['_id'] = saved['id']

    @coroutine
    def save_subresource(self, subresource):
        """
        Save the sub-resource

        NOTE: Currently assumes subresources are stored within a dictionary,
        keyed with the subresource's ID
        """
        data = deepcopy(subresource._resource)
        data.pop('id', None)
        data.pop(self.resource_type + '_id', None)

        subresources = getattr(self, subresource.parent_key, {})
        subresources[subresource.id] = data
        setattr(self, subresource.parent_key, subresources)

        yield self._save()

    @coroutine
    def delete(self, user):
        """Delete a resource"""
        if user:
            can_delete = yield self.can_delete(user)
        else:
            can_delete = False

        if not can_delete:
            raise exceptions.Unauthorized('User may not delete the resource')

        doc = {
            '_id': self.id,
            '_deleted': True
        }

        try:
            doc['_rev'] = self._rev
        except AttributeError:
            pass

        db = self.db_client()
        yield db.save_doc(doc)

        self._resource = doc

    @coroutine
    def delete_subresource(self, subresource):
        """
        Delete the sub-resource

        NOTE: Currently assumes subresources are stored within a dictionary,
        keyed with the subresource's ID
        """
        subresources = getattr(self, subresource.parent_key, {})
        del subresources[subresource.id]
        yield self._save()

    def clean(self):
        """Remove internal fields"""
        doc = self._resource
        result = {k: v for k, v in doc.iteritems() if k not in
                  self.internal_fields}

        if '_id' in doc and 'id' not in result:
            result['id'] = doc['_id']

        return result

    @classmethod
    @coroutine
    def can_create(cls, user, **data):
        """Check if a user is authorized to create a resource"""
        raise Return(True)

    @coroutine
    def can_approve(self, user, **data):
        """Check if a user is authorized to approve a resource"""
        raise Return(False)

    @coroutine
    def can_update(self, user, **data):
        """Check if a user is authorized to update a resource"""
        raise Return((True, set([])))

    @coroutine
    def can_read(self, user):
        """Check if a user is authorized to read a resource"""
        raise Return(True)

    @coroutine
    def can_delete(self, user):
        """Check if a user is authorized to delete a resource"""
        raise Return(False)

    @property
    def state(self):
        """Get the Document's state"""
        state = self._resource.get('state', self.default_state)

        if state in State:
            return state
        else:
            return getattr(State, state)

    @state.setter
    def state(self, value):
        self._resource['state'] = value


class SubResource(Document):
    _parent = None
    schema = Schema({'id': unicode}, extra=ALLOW_EXTRA)

    def __init__(self, parent=None, **kwargs):
        self._resource = kwargs
        if parent:
            self._parent = parent

    @property
    def _read_only(self):
        return set(self.read_only_fields) | {'id', 'type'}

    @property
    def parent_id(self):
        return getattr(self, self.parent_resource.resource_type + '_id')

    @property
    def parent(self):
        return self._parent

    @coroutine
    def get_parent(self):
        """
        Get the parent resource from the database

        The get, create & update methods will populate the parent for you. Use
        this method in the cases where parent has not been populated.
        """
        if not self._parent:
            self._parent = yield self.parent_resource.get(self.parent_id)

        raise Return(self._parent)

    @classmethod
    def parent_resources(cls):
        """Get a list of parent resources, starting from the Document"""
        parent = cls.parent_resource
        parents = [parent]

        try:
            while True:
                parent = parent.parent_resource
                parents.append(parent)
        except AttributeError:
            pass

        parents.reverse()
        return parents

    @classmethod
    def parent_document(cls):
        """
        Walk up all parent_resources to get the Document

        Might not be the same as self.parent_resource because could be a
        sub-sub-resource
        """
        return cls.parent_resources()[0]

    @property
    def id(self):
        try:
            return self._resource['id']
        except KeyError:
            raise AttributeError('id')

    @id.setter
    def id(self, value):
        self._resource['id'] = value

    @classmethod
    @coroutine
    def create(cls, user, **kwargs):
        """If parent resource is not an editable state, should not be able to create."""
        parent_id = kwargs.get(cls.parent_resource.resource_type + '_id')
        try:
            parent = yield cls.parent_resource.get(parent_id)
        except couch.NotFound:
            msg = 'Parent {} with id {} not found'.format(
                cls.parent_resource.resource_type,
                parent_id)
            raise exceptions.ValidationError(msg)

        if not parent.editable:
            err = 'Cannot create child of {} resource'.format(parent.state.name)
            raise exceptions.Unauthorized(err)

        resource = yield super(SubResource, cls).create(user, **kwargs)
        resource._parent = parent

        raise Return(resource)

    @coroutine
    def update(self, user, **kwargs):
        """If parent resource is not an editable state, should not be able to update"""
        yield self.get_parent()

        if not self.parent.editable:
            err = 'Cannot update child of {} resource'.format(self.parent.state.name)
            raise exceptions.Unauthorized(err)

        yield super(SubResource, self).update(user, **kwargs)

    @coroutine
    def _save(self):
        """Save the sub-resource within the parent resource"""
        yield self.validate()

        try:
            self._parent = yield self.parent_resource.get(self.parent_id)
        except couch.NotFound:
            msg = '{}_id {} not found'.format(
                self.parent_resource.resource_type,
                self.parent_id)
            raise exceptions.ValidationError(msg)

        yield self._parent.save_subresource(self)

    @classmethod
    @coroutine
    def get(cls, resource_id, include_deactivated=False):
        """
        Get a resource

        :param resource_id: the resource ID
        :param include_deactivated: Include deactivated resources in response
        :returns: a SubResource instance
        :raises: SocketError, CouchException
        """
        if include_deactivated:
            resource = yield cls.view.first(key=resource_id, include_docs=True)
        else:
            resource = yield cls.active_view.first(key=resource_id, include_docs=True)
        parent = cls.parent_resource(**resource['doc'])

        raise Return(cls(parent=parent, **resource['value']))

    @classmethod
    @coroutine
    def all(cls, include_deactivated=False):
        """
        Get all sub-resources
        :param include_deactivated: Include deactivated resources in response

        :returns: list of SubResource instances
        :raises: SocketError, CouchException
        """
        if include_deactivated:
            resources = yield cls.view.get(include_docs=True)
        else:
            resources = yield cls.active_view.get(include_docs=True)

        result = []

        for resource in resources['rows']:
            parent = cls.parent_resource(**resource['doc'])
            result.append(cls(parent=parent, **resource['value']))

        raise Return(result)

    @coroutine
    def delete(self, user):
        """Delete a sub-resource"""
        if user:
            can_delete = yield self.can_delete(user)
        else:
            can_delete = False

        if not can_delete:
            raise exceptions.Unauthorized('User may not delete the resource')

        try:
            parent = yield self.get_parent()
        except couch.NotFound:
            msg = '{}_id {} not found'.format(
                self.parent_resource.resource_type,
                self.parent_id)
            raise exceptions.ValidationError(msg)

        yield parent.delete_subresource(self)

    @property
    def state(self):
        """
        Get the SubResource state

        If the parents state has a higher priority, then it overrides the
        SubResource state

        ..note:: This assumes that self.parent is populated
        """
        state = self._resource.get('state', self.default_state)
        if state not in State:
            state = getattr(State, state)

        if not self.parent:
            raise Exception('Unable to check the parent state')

        parent_state = self.parent.state

        return max([state, parent_state], key=attrgetter('value'))



# -*- coding: utf-8 -*-
# Copyright 2016 Open Permissions Platform Coalition
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.

import couch
from tornado import httpclient


class ValidationError(Exception):
    pass


class Unauthorized(Exception):
    pass


class NotFound(couch.NotFound):
    def __init__(self):
        http_error = httpclient.HTTPError(404)
        couch.NotFound.__init__(self, http_error)



# -*- coding: utf-8 -*-
# Copyright 2016 Open Permissions Platform Coalition
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.

from setuptools import setup
import re

with open('perch/__init__.py', 'r') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError('Cannot find version information')

setup(
    name='opp-perch',
    version=version,
    description='Open Permissions Platform Couchdb Library',
    author='Open Permissions Platform Coalition',
    author_email='support@openpermissions.org',
    url='https://github.com/openpermissions/perch',
    packages=['perch', 'perch.migrations'],
    entry_points='''
        [console_scripts]
        perch=perch.cli:cli
    ''',
    install_requires=["tornado-couchdb==0.2.3",
                      "enum34==1.0.4",
                      "passlib==1.6.2",
                      "python-dateutil==2.5.2",
                      "click",
                      "voluptuous==0.8.9"]
)




from __future__ import unicode_literals
import perch

VERSION = '22e76f4ff8bd41e19aa52839fc8f13a1'
PREVIOUS_VERSION = ''


@perch.migrate.migration(perch.User, VERSION, PREVIOUS_VERSION)
def migrate_user(instance):
    """
    Move User.organisations['global']['role'] to top-level property and remove
    verified flag
    """
    instance._resource.pop('verified', None)

    if 'role' in instance._resource:
        return instance

    global_org = instance.organisations.pop('global', {})
    instance.role = global_org.get('role', perch.User.roles.default.value)

    return instance



# -*- coding: utf-8 -*-
# Copyright 2016 Open Permissions Platform Coalition
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations under
# the License.

import inspect
from collections import defaultdict
from functools import partial

from tornado.gen import coroutine, Return
from tornado.options import options
import couch

from . import exceptions


_views = defaultdict(list)


class View(object):

    def __init__(self, db_name, func):
        """
        Wraps a map function, adding a get method.
        :param: db_name: name of database
        :param: func: map function
        """
        self.db_name = db_name
        self.func = func

    @property
    def name(self):
        return self.func.func_name

    def __call__(self, doc):
        return self.func(doc)

    def register(self):
        """Register the view so that can be loaded into Couch"""
        _views[self.db_name].append(self.create_design_doc())

    def create_design_doc(self):
        """Create a design document from a Python map function"""
        source = [x for x in inspect.getsourcelines(self.func)[0]
                  if not x.startswith('@')]

        doc = {
            '_id': '_design/{}'.format(self.name),
            'language': 'python',
            'views': {
                self.name: {
                    'map': ''.join(source)
                }
            }
        }

        return doc

    @coroutine
    def get(self, **kwargs):
        """
        Queries database for results of view
        :return:
        """
        db_url = ':'.join([options.url_registry_db, str(options.db_port)])
        db = couch.AsyncCouch(db_name=self.db_name, couch_url=db_url)

        result = yield db.view(
            design_doc_name=self.name,
            view_name=self.name,
            **kwargs)

        raise Return(result)

    @coroutine
    def first(self, **kwargs):
        """
        Queries database for first result of view
        :return:
        """
        result = yield self.get(**kwargs)

        if not result['rows']:
            raise exceptions.NotFound()

        raise Return(result['rows'][0])

    @coroutine
    def values(self, **kwargs):
        """Get the view's values"""
        result = yield self.get(**kwargs)

        if not result['rows']:
            raise Return([])

        raise Return([x['value'] for x in result['rows']])


def view(db_name):
    """
    Register a map function as a view

    Currently, only a single map function can be created for each view

    NOTE: the map function source is saved in CouchDB, so it cannot depend on
    anything outside the function's scope.

    :param db_name: the database name
    """
    def decorator(func):
        v = View(db_name, func)
        v.register()
        return v
    return decorator


def load_design_docs():
    """
    Load design docs for registered views
    """
    url = ':'.join([options.url_registry_db, str(options.db_port)])
    client = partial(couch.BlockingCouch, couch_url=url)

    for name, docs in _views.items():
        db = client(db_name=name)
        views = []

        for doc in docs:
            try:
                current_doc = db.get_doc(doc['_id'])

                # use the current _rev if not provided
                if '_rev' not in doc:
                    doc['_rev'] = current_doc['_rev']
            except couch.NotFound:
                pass

            views.append(doc)

        db.save_docs(views)


@view('registry')
def resource_version(doc):
    """View for querying resources by version (used for data migrations)"""
    yield (doc.get('type'), doc.get('doc_version', '')), doc['_id']


@view('registry')
def active_users(doc):
    """View for getting users"""
    if doc.get('type') == 'user' and doc.get('state') != 'deactivated':
        yield doc.get('email'), doc['_id']


@view('registry')
def users(doc):
    """View for getting users"""
    if doc.get('type') == 'user':
        yield doc.get('email'), doc['_id']


@view('registry')
def active_user_organisations_resource(doc):
    """Get user.organisations subresouces"""
    if doc.get('type') == 'user' and doc.get('state') != 'deactivated':
        for org_id, resource in doc.get('organisations', {}).items():
            if resource['state'] != 'deactivated':
                resource['id'] = org_id
                resource['user_id'] = doc['_id']
                yield [doc['_id'], org_id], resource


@view('registry')
def user_organisations_resource(doc):
    """Get user.organisations subresouces"""
    if doc.get('type') == 'user':
        for org_id, resource in doc.get('organisations', {}).items():
            resource['id'] = org_id
            resource['user_id'] = doc['_id']
            yield [doc['_id'], org_id], resource


@view('registry')
def active_joined_organisations(doc):
    """View for getting organisations associated with a user"""
    if doc.get('type') == 'user' and doc.get('state') != 'deactivated':
        for org_id, state in doc.get('organisations', {}).items():
            if state['state'] == 'deactivated':
                continue

            org = {'_id': org_id}
            yield [doc['_id'], None], org

            try:
                yield [doc['_id'], state['state']], org
            except KeyError:
                pass


@view('registry')
def joined_organisations(doc):
    """View for getting organisations associated with a user"""
    if doc.get('type') == 'user':
        for org_id, state in doc.get('organisations', {}).items():
            org = {'_id': org_id}
            yield [doc['_id'], None], org

            try:
                yield [doc['_id'], state['state']], org
            except KeyError:
                pass


@view('registry')
def organisation_members(doc):
    """Lookup users that have joined an organisation"""
    if doc.get('type') == 'user':
        for org_id in doc.get('organisations', {}):
            yield org_id, doc['_id']


@view('registry')
def admin_emails(doc):
    """View for an orginsation's admin email addresses"""
    if doc.get('type') == 'user' and doc.get('state') != 'deactivated':
        for org_id, state in doc.get('organisations', {}).items():
            if state.get('role') == 'administrator' and state.get('state') != 'deactivated':
                yield org_id, doc['email']


@view('registry')
def active_organisations(doc):
    """View for getting active organisations"""
    if doc.get('type') == 'organisation' and doc.get('state') != 'deactivated':
        yield doc.get('state'), doc['_id']


@view('registry')
def organisations(doc):
    """View for getting organisations"""
    if doc.get('type') == 'organisation':
        yield doc.get('state'), doc['_id']


@view('registry')
def organisation_name(doc):
    """View for getting organisations by their name"""
    if doc.get('type') == 'organisation':
        yield doc.get('name'), doc['_id']


@view('registry')
def reference_links(doc):
    """Get reference links"""
    if doc.get('type') == 'organisation' and doc.get('state') != 'deactivated':
        for asset_id_type, link in doc.get('reference_links', {}).get('links', {}).items():
            value = {
                'organisation_id': doc['_id'],
                'link': link
            }
            yield asset_id_type, value


@view('registry')
def active_services(doc):
    """View for getting active services"""
    if doc.get('state') != 'deactivated':
        for service_id, service in doc.get('services', {}).items():
            if service.get('state') != 'deactivated':
                service_type = service.get('service_type')
                org = doc['_id']
                service['id'] = service_id
                service['organisation_id'] = org

                yield service_id, service
                yield [service_type, org], service
                yield [service_type, None], service
                yield [None, org], service
                yield [None, None], service


@view('registry')
def services(doc):
    """View for getting services"""
    for service_id, service in doc.get('services', {}).items():
        service_type = service.get('service_type')
        org = doc['_id']
        service['id'] = service_id
        service['organisation_id'] = org

        yield service_id, service
        yield [service_type, org], service
        yield [service_type, None], service
        yield [None, org], service
        yield [None, None], service


@view('registry')
def active_service_location(doc):
    """View for getting active service by location"""
    if doc.get('state') != 'deactivated':
        for service_id, service in doc.get('services', {}).items():
            if service.get('state') != 'deactivated':
                service['id'] = service_id
                service['organisation_id'] = doc['_id']

                location = service.get('location', None)
                if location:
                    yield location, service


@view('registry')
def service_location(doc):
    """View for getting service by location"""
    for service_id, service in doc.get('services', {}).items():
        service['id'] = service_id
        service['organisation_id'] = doc['_id']

        location = service.get('location', None)
        if location:
            yield location, service


@view('registry')
def service_name(doc):
    """View for getting service by name"""
    for service_id, service in doc.get('services', {}).items():
        service['id'] = service_id
        service['organisation_id'] = doc['_id']

        name = service.get('name', None)
        if name:
            yield name, service


@view('registry')
def oauth_secrets(doc):
    """View for OAuth secrets"""
    if doc.get('type') == 'oauth_client_credentials':
        yield doc.get('client_id'), doc['_id']


@view('registry')
def oauth_client(doc):
    """View for getting OAuth clients using the secret"""
    if doc.get('type') == 'oauth_client_credentials':
        yield [doc['_id'], doc['client_id']], doc['_id']


@view('registry')
def active_repositories(doc):
    """View for getting active repositories"""
    if doc.get('state') != 'deactivated':
        for repository_id, repo in doc.get('repositories', {}).items():
            if repo.get('state') != 'deactivated':
                repo['id'] = repository_id
                repo['organisation_id'] = doc['_id']

                yield repository_id, repo


@view('registry')
def repositories(doc):
    """View for getting repositories"""
    for repository_id, repo in doc.get('repositories', {}).items():
        repo['id'] = repository_id
        repo['organisation_id'] = doc['_id']

        yield repository_id, repo


@view('registry')
def repository_name(doc):
    """View for checking repository name is unique"""
    for repository_id, repo in doc.get('repositories', {}).items():
        repo['id'] = repository_id
        repo['organisation_id'] = doc['_id']

        name = repo.get('name', None)
        if name:
            yield name, repository_id


@view('registry')
def service_and_repository(doc):
    """
    View for looking up services and repositories by their ID

    Used in the auth service
    """
    if doc.get('type') == 'organisation' and doc.get('state') != 'deactivated':
        for repository_id, repo in doc.get('repositories', {}).items():
            if repo.get('state') != 'deactivated':
                repo['id'] = repository_id
                repo['organisation_id'] = doc['_id']

                yield repository_id, repo

        for service_id, service in doc.get('services', {}).items():
            if service.get('state') != 'deactivated':
                service['id'] = service_id
                service['organisation_id'] = doc['_id']

                yield service_id, service
